import sys
from typing import List, TextIO
from abc import ABC, abstractmethod
import os
import datetime
from views_schema import viewser as schema
from pymonad.maybe import Nothing
from viewser.tui.formatting import errors as error_formatting

class ErrorDumpIO(ABC):
    """
    ErrorDumpIO
    ===========

    Abstract class for error handling, outputting error content (schema.Dump)
    to some output.
    """

    @abstractmethod
    def write(self, dump: schema.Dump):
        pass

    def strftime(self, time: datetime.datetime):
        return time.strftime("%Y-%m-%d_%H:%M:%S")


class FileErrorHandler(ErrorDumpIO):
    """
    FileErrorHandler
    ================

    Dumps error contents to file.
    """

    def __init__(self, directory: str):
        self._directory = directory
        os.makedirs(self._directory, exist_ok = True)

    def write(self, dump: schema.Dump):
        with open(self._path(dump), "w") as f:
            f.write(dump.json())

    def _name(self, dump: schema.Dump):
        return f"viewser_{self.strftime(dump.timestamp)}_{dump.title}.json"

    def _path(self, dump: schema.Dump):
        return os.path.join(self._directory, self._name(dump))

class StreamHandler(ErrorDumpIO):
    """
    StreamHandler
    ==================

    Reports back errors to the user via a stream (default sys.stdout).
    """

    def __init__(self, stream: TextIO = sys.stdout):
        self._stream: TextIO                 = stream

        self._buffer: str                    = ""
        self._hints: List[schema.Message]    = []
        self._messages: List[schema.Message] = []

        self._formatter                      = error_formatting.ErrorFormatter()

        super().__init__()

    def write(self, dump: schema.Dump):
        self._stream.write(self._formatter.formatted(dump))

class ErrorDumper():
    def __init__(self, writers: List[ErrorDumpIO] = None):
        if writers is None:
            writers = [StreamHandler()]
        self._writers = writers

    def dump(self, error_dump: schema.Dump) -> Nothing:
        for writer in self._writers:
            writer.write(error_dump)
        return Nothing



from sys import platform
import time
import webbrowser
import json
import hashlib
from urllib.parse import urlencode,urlunparse
from operator import add
import random
import logging
from typing import Callable, Set, Tuple, Optional, Dict
from contextlib import closing
import os
import requests
import pydantic
from toolz.functoolz import reduce, curry, do
from pymonad.maybe import Maybe
from pymonad.either import Either, Left, Right
import docker
import psutil
from viewser import settings
from viewser.error_handling import exceptions
from viewser.tui import ascii_art

START_PORT = 2000
MAX_PORT=9999

logger = logging.getLogger(__name__)


digest = lambda s: hashlib.md5(s.encode()).hexdigest()
generate_token = lambda: digest(reduce(add, [str(random.random()) for i in range(128)]))

if platform in ["linux","linux2"]:
    occupied_ports:Callable[[], Set[int]] = lambda: {c.laddr.port for c in psutil.net_connections()}
else:
    occupied_ports = set

def seek_port(from_port:int) -> Optional[int]:
    if from_port > MAX_PORT:
        return None
    else:
        return from_port if from_port not in occupied_ports() else seek_port(from_port+1)

def make_url(port:int, token:str, host: str = "0.0.0.0"):
    return urlunparse(("http", f"{host}:{port}","","",urlencode({"token":token}),""))

class ServicePrincipal(pydantic.BaseModel):
    clientId: str
    clientSecret: str

    def login_info(self):
        return {
                "username": self.clientId,
                "password": self.clientSecret
            }

def local_sp() -> Either[str, ServicePrincipal]:
    try:
        sp_file_path = os.path.join(settings.CONFIG_DIR, "sp.json")
        with open(sp_file_path) as f:
            return Right(ServicePrincipal(**json.load(f)))
    except FileNotFoundError:
        return Left(f"Could not find SP file at {sp_file_path}")
    except pydantic.ValidationError:
        return Left(f"SP file at {sp_file_path} is corrupted")

def assure_logged_in(
        client: docker.DockerClient,
        registry:str,
        service_principal: ServicePrincipal)-> Dict[str,str]:
    try:
        logger.info(f"Authenticating to registry {registry} using local credentials")
        return Right(client.login(
                    **service_principal.login_info(),
                    registry = registry))
    except docker.errors.APIError as apie:
        return Left(
                f"Failed to authenticate to registry {registry}.\n"
                "Bad credentials?\n"
                f"Raw error: \"{str(apie.explanation)}\"")

def notebook_image(
        registry: str,
        repository: str,
        version: str,
        pull: bool) -> Either[str,str]:

    name = registry + "/" + repository

    if pull:
        logger.info("Refreshing docker image (this might take a while...)")
        with closing(docker.client.from_env()) as client:
            return (local_sp()
                .then(curry(assure_logged_in,client,registry))
                .then(lambda _: Right(client.images.pull(name, tag = version)))
                .then(curry(do,lambda img: logger.info(f"Using {img.short_id}")))
                .then(lambda img: Right(img.tags[0]))
                )
    else:
        tagged_name = name + ":" + version
        return Right(tagged_name)

def run_notebook_server(
        port: int,
        working_directory: str,
        requirements_file: Maybe[str],
        image: str,
        )-> Tuple[str, str]:

    volumes = {
        os.path.abspath(working_directory):{
            "bind":"/home/views/notebooks",
            "mode": "rw"
            },
        os.path.expanduser("~/.views/config.json"):{
            "bind": "/home/views/.views/config.json",
            "mode": "ro"
            }
        }

    volumes = requirements_file.maybe(volumes, lambda fname: {
            os.path.abspath(fname): {
                "bind": "/home/views/user_requirements.txt",
                "mode": "ro"
                }, **volumes
        })

    token = generate_token()

    with closing(docker.client.from_env()) as client:
        configuration = {
                "image": image,
                "remove": True,
                "detach": True,
                "volumes" : volumes,
                "environment":{
                    "NOTEBOOK_SERVER_HOST":"0.0.0.0",
                    "NOTEBOOK_SERVER_TOKEN": token,
                    },
                "ports": {
                    "8888/tcp": port
                    }
            }

        container = client.containers.run(**configuration)
        return (container.id, make_url(port, token))

def watch(browser: bool, echo: Callable[[str],None], run_report: Tuple[str, str])-> None:
    container_id, container_url = run_report
    echo("Booting up...")

    status_code = -1
    with closing(docker.client.from_env()) as docker_client:
        container = docker_client.containers.get(container_id)
        for line in container.attach(stream = True):
            if status_code != 200:
                print(line.decode())
                try:
                    status_code = requests.get(container_url).status_code
                except requests.ConnectionError:
                    status_code = -1
            else:
                break

        echo("Made connection, checking container status...")
        try:
            time.sleep(1)
            print(docker_client.containers.get(container_id))
        except docker.errors.NotFound:
            raise exceptions.ViewserspaceError(
                    "Container failed to launch",
                    hint = (
                        "This was probably due to a failed dependency download. "
                        "Check the logs above."
                        )
                    )
        else:
            echo("Container OK!")

    echo(ascii_art.VIEWSERSPACE_LOGO)
    echo("Server running. Ctrl-C to stop.")

    if browser:
        webbrowser.open(container_url)
    else:
        echo(f"To access, open {container_url}")

    try:
        while True:
            pass
    finally:
        container = docker.client.from_env().containers.get(container_id)
        container.kill()




import os
from typing import List
from pydantic import BaseModel

class ErrorDumpFile(BaseModel):
    name: str
    size: int

class ErrorDumpFiles(BaseModel):
    directory: str
    files:     List[ErrorDumpFile]

    @classmethod
    def from_dir(cls: "ErrorDumpFiles", dir: str):
        size = lambda f: os.path.getsize(os.path.join(dir, f))
        files = [ErrorDumpFile(name = f, size = size(f)) for f in os.listdir(dir)]
        return cls(directory = dir, files = files)



from collections import defaultdict
from typing import Optional
import json
import functools

import pkg_resources
import colorama
import click

def raise_pretty_exception(x: Exception):
    try:
        raise x
    except OperationPending:
        raise x
    finally:
        try:
            x.is_pretty
        except AttributeError:
            raise ViewsError(message = str(x)) from x
        else:
            raise x

def with_ansi(ansi):
    def wrapper(fn):
        @functools.wraps(fn)
        def inner(self,msg,*args,**kwargs):
            msg = ansi + msg + colorama.Style.RESET_ALL
            fn(self,msg,*args,**kwargs)
        return inner
    return wrapper

class PrettyFormatter(click.HelpFormatter):
    def __init__(self,*args,**kwargs):
        colorama.init()
        super().__init__(*args,**kwargs)

    @with_ansi(colorama.Fore.RED + colorama.Style.BRIGHT)
    def write_heading(self, heading):
        super().write_heading(heading)

class PrettyError(click.ClickException):
    error_name = "Error"
    is_pretty = True
    DEFAULT_HINT = None

    def __init__(self, message: str, hint: Optional[str] = None):
        hint = hint if hint else (self.DEFAULT_HINT if self.DEFAULT_HINT else None)
        super().__init__(self.pretty_format(message,hint))

    def pretty_format(self, message: str, hint: Optional[str] = None):
        formatter = PrettyFormatter()
        formatter.write_heading(self.error_name)
        formatter.indent()
        if message:
            with formatter.section("Description"):
                formatter.write_text(message)

        if hint:
            with formatter.section("Hint"):
                formatter.write_text(colorama.Fore.GREEN + hint + colorama.Style.RESET_ALL)

        return formatter.getvalue()

class ViewsError(PrettyError):
    error_name = "ViEWS error"

class ExistsError(PrettyError):
    """
    Raised when something is in conflict with remote,
    and an override has not been passed (--override).
    """
    error_name = "Already exists!"
    def __init__(self):
        super().__init__(self.pretty_format(
            "This resource already exists!",
            "Try passing --overwrite if you want to overwrite this resource"
            ))

class ConfigurationError(PrettyError):
    """
    Raised when something is (assumed to be) misconfigured
    """
    error_name = "Configuration error"

    def __init__(self,message: str, hint: Optional[str] = None):
        super().__init__(self.pretty_format(message, hint))

class ConnectionError(PrettyError):
    error_name = "Connection error"

    def __init__(self, url):
        super().__init__(
                    f"Could not connect to {url}",
                    (
                        "Is the config setting REMOTE_URL set to the right value?\n\n"
                        "To check, run "
                        "\n\nviewser config get REMOTE_URL.\n\n"
                        "To change the value, run "
                        "\n\nviewser config set REMOTE_URL\n\n"
                    )
                )

class UrlParsingError(PrettyError):
    error_name = "URL parsing error"
    def __init__(self, url):
        super().__init__(
                    f"Could not parse {url} as a valid http(s) url",
                    (
                        "Set the config setting REMOTE_URL to a valid http url:\n\n"
                        "viewser config set REMOTE_URL {a valid url}.\n\n"
                        "Did you forget to add http(s):// to the url?"
                    )
                )

class DeserializationError(PrettyError):
    error_name = "DeserializationError"

    def __init__(self, bytes):
        super().__init__(f"""
            Could not deserialize as parquet:
            \"{bytes[:50]}{'...' if len(bytes)>50 else ''}\"
        """)

class RequestError(PrettyError):
    error_name = "Request error"

    def _http_default_hints(self):
        return {}

    def http_defaults(self, status_code):
        return defaultdict(str)

    def make_message(self):
        return self.content

    def __init__(self, response, hint: Optional[str] = None):
        self.response = response
        self.url = response.url
        self.status_code = response.status_code

        defaults = self.http_defaults(self.status_code)

        self.content = response.content.decode().strip() if response.content else defaults["phrase"]
        self.hint = hint if hint is not None else defaults["description"]
        super().__init__(self.content, self.hint)

class OperationPending(RequestError):
    error_name = "Pending..."

class ClientError(RequestError):
    error_name = "4xx client error"

class NotFoundError(RequestError):
    error_name = "404 not found"

class RemoteError(RequestError):
    error_name = "5xx remote error"

class RequestAssertionError(RequestError):
    error_name = "Assertion error"
    def __init__(self, attr, expected, response):
        self.attr = attr
        self.expected = expected
        self.actual = getattr(attr, response)
        super().__init__(response)

    def make_message(self):
        return (
                f"Response {self.attr} had unexpected value: "
                f"Expected {self.expected}, got {self.actual}"
            )

class ViewserspaceError(PrettyError):
    error_name = "Viewserspace error"
    DEFAULT_HINT = """
These errors are usually related to Docker / Azure ACR authentication. Make
sure that you have installed the credentials file in the proper location
(~/.views/sp.json), and that the credentials are correct. Talk to an admin for
credentials.
    """

    def __init__(self, message, hint: Optional[str] = None):
        hint = hint if hint is not None else self.DEFAULT_HINT
        super().__init__(message, hint)

def exception_raiser(exception:Exception, *args, **kwargs):
    e = exception(*args, **kwargs)
    raise e

class MaxRetries(PrettyError):
    error_name = "Max retries"

    DEFAULT_HINT = """
Your request reached the currently configured maximum number of retries.
To increase the limit for max retries, run this command:

viewser config set MAX_RETRIES { a bigger number }
    """

    def __init__(self):
        super().__init__("Max number of retries reached")



import subprocess
from typing import Callable
from importlib.metadata import version
import re
import json
import logging
import functools
import requests

from toolz.functoolz import curry, compose
from pymonad.maybe import Just,Nothing,Maybe
from pymonad.either import Either, Left, Right
from . import remotes

logger = logging.getLogger(__name__)

async def log_after(function):
    result = await function()
    result.then(logger.warning)

def maybe_log(check_fn: Callable[[], Maybe[str]], fn):
    """
    Function decorator that might emit a warning log message, if test_fn
    returns Just(str)

    Arguments are passed to test_fn
    """
    @functools.wraps(fn)
    def inner(*args,**kwargs):
        check_fn().then(logger.warning)
        return fn(*args,**kwargs)
    return inner

async def check_server_version(url)-> Maybe[str]:
    """
    Used to decorate operations that connect to remote API(s).  The remotes
    should define a "handshake" path, which should return json containing a
    "viewser_version" field.
    """
    add_header = lambda x: "Error while checking remote version\n" + x

    def check(response: requests.Response):
        try:
            remote_version = response.json()["viewser_version"]
        except (json.JSONDecodeError,KeyError):
            return Just("The handshake endpoint did not return the right data. "
                    "Expected JSON  wiith a 'viewser_version' key, got "
                    f"{response.content.decode()}"
                    )

        try:
            assert re.search(r"[0-9]+\.[0-9]+\.[0-9]+",remote_version)
        except AssertionError:
            return Just("Handshake version wrong format: {remote_version} (expected n.n.n)")

        major_version = lambda x: x.split(".")[0]

        logger.debug("%s vs %s",remote_version, version("viewser"))
        if not major_version(remote_version) == major_version(version("viewser")):
            return Just(
                    "Viewser installation has wrong major version. "
                    f"Local is {version('viewser')} while expected is {remote_version}. "
                    "Please install the correct version by running "
                    f"pip install --upgrade viewser~={remote_version}"
                )
        return Nothing

    return (remotes.request(
            url,
            "GET",
            [curry(remotes.check_content_type,"application/json")] + remotes.status_checks,
            "")
            .either(
                compose(Just,add_header,str),
                compose(lambda m: m.then(add_header),check)
                )
        )

def get_pypi_version() -> Either[str,str]:
    with subprocess.Popen(["pip","show","viewser"],
            stdout = subprocess.PIPE,
            stderr = subprocess.PIPE) as proc:
        out,err = proc.communicate()
        if not err:
            version_line,*_ = [l for l in out.split(b"\n") if b"Version" in l]
            return Right(version_line.split()[-1].decode())
        else:
            return Left(err)

def maybe_version_number(raw: str)-> Either[str,str]:
    if re.search(r"[0-9]+\.[0-9]+\.[0-9]+",raw):
        return Right(raw)
    else:
        return Left(f"{raw} is not a version number")

def check_pypi_version() -> Maybe[str]:
    """
    Gets the latest version of the CLI from Github
    """
    logger.debug("checking viewser version on PyPi")
    def version_check(latest)->Maybe[str]:
        current = version("viewser")
        if not latest == current:
            return Just(
                    "There is a newer version of viewser available! "
                    f"You are currently on {current}, but the "
                    f"latest version is {latest}. Please update by "
                    "running pip install --upgrade viewser."
                )
        return Nothing

    return (get_pypi_version()
        .either(compose(Just,str), version_check)
        )




import io
from typing import Dict, Any, Optional
import click
from views_schema import docs as schema
from viewser import settings
from . import formatting, operations

@click.group(name = "tables", short_help = "show information about available tables")
@click.pass_obj
def tables_cli(obj: Dict[str, Any]):
    """
    Commands used to inspect available tables and columns.
    """
    obj["operations"]       = operations.DocumentationCrudOperations(settings.config_get("REMOTE_URL"), "docs/tables")
    obj["detail_formatter"] = formatting.DocumentationDetailFormatter()
    obj["list_formatter"]   = formatting.DocumentationTableFormatter()

@tables_cli.command(name="list", short_help="show available tables")
@click.pass_obj
def list_tables(obj: Dict[str, Any]):
    """
    list_tables
    ===========

    Show all available tables.
    """

    click.echo(obj["operations"].list()
        .either(obj["error_dumper"].formatted, obj["list_formatter"].formatted))

@tables_cli.command(name="show", short_help="inspect table or column")
@click.argument("table-name")
@click.argument("column-name", required=False)
@click.pass_obj
def show_tables(
        obj: Dict[str, Any],
        table_name: str,
        column_name: Optional[str] = None):
    """
    Show either a table or a column, depending on whether one argument is
    passed (table name), or two arguments are passed (table name - column
    name).
    """
    path = table_name

    if column_name:
        path += "/"+column_name
        formatter = obj["detail_formatter"]
    else:
        formatter = obj["list_formatter"]

    click.echo(obj["operations"].show(path)
            .either(obj["error_dumper"].formatted, formatter.formatted))

@tables_cli.command(name="annotate", short_help="add documentation text")
@click.argument("content-file", type=click.File("r"))
@click.argument("table-name")
@click.argument("column-name", required=False)
@click.option("--overwrite", is_flag = True, help="overwrite existing documentation")
@click.pass_obj
def annotate_table(
        obj: Dict[str, Any],
        content_file: io.BufferedReader,
        table_name: str,
        column_name: str,
        overwrite: bool):
    """
    Annotate a table or a column. To annotate a table, pass a single argument
    following the file containing the annotation. To annotate a column, pass
    two arguments.
    """

    path = table_name
    if column_name:
        path += "/"+column_name
        formatter = obj["detail_formatter"]
    else:
        formatter = obj["list_formatter"]

    to_post = schema.PostedDocumentationPage(content = content_file.read())

    click.echo(obj["operations"].post(to_post, path, overwrite = overwrite)
        .either(obj["error_dumper"].formatted, formatter.formatted))

@click.group(name="transforms")
@click.pass_obj
def transforms_cli(obj: Dict[str, Any]):
    """
    Commands used to inspect available transforms.
    """

    obj["operations"] = operations.DocumentationCrudOperations(settings.config_get("REMOTE_URL"), "docs/transforms")
    obj["table_formatter"] = formatting.DocumentationTableFormatter()
    obj["detail_formatter"] = formatting.FunctionDetailFormatter()

@transforms_cli.command(name="list", short_help="show all available transforms")
@click.pass_obj
def list_transforms(obj: Dict[str, Any]):
    """
    List all available transforms.
    """

    click.echo(obj["operations"].list()
        .either(obj["error_dumper"].formatted, obj["table_formatter"].formatted))

@transforms_cli.command(name="show",short_help="show details about a transform")
@click.argument("transform-name", type=str)
@click.pass_obj
def show_transform(
        obj: Dict[str, Any],
        transform_name: str):
    """
    Show details about a transform, such as which arguments it takes, and
    which level of analysis it is applicable to.
    """
    click.echo(obj["operations"].show(transform_name)
        .either(obj["error_dumper"].formatted, obj["detail_formatter"].formatted))

@transforms_cli.command(name="annotate", short_help="add documentation text")
@click.argument("content-file", type=click.File("r"))
@click.argument("transform-name")
@click.option("--overwrite", is_flag = True, help="overwrite existing documentation")
@click.pass_obj
def annotate_transform(
        obj: Dict[str, Any],
        content_file: io.BufferedReader,
        transform_name: str,
        overwrite: bool = False):
    """
    Annotate a table or a column. To annotate a table, pass a single argument
    following the file containing the annotation. To annotate a column, pass
    two arguments.
    """
    to_post = schema.PostedDocumentationPage(content = content_file.read())
    click.echo(obj["operations"].post(to_post, transform_name, overwrite=overwrite)
        .either(obj["error_dumper"].formatted, obj["detail_formatter"].formatted))



from importlib.metadata import version
import click
from viewser import settings
from . import operations

@click.group(name="help", short_help="commands for finding documentation")
def cli():
    """
    Commands related to finding documentation.
    """

@cli.command(name = "wiki")
def wiki():
    """
    Open the viewser wiki in a new browser window, which contains all you need
    to know about using viewser.
    """
    operations.open_browser(settings.config_get("REPO_URL"),"wiki")

@cli.command(name = "issue")
def issue():
    """
    Opens the viewser "new issue" page in a new browser window.
    """
    operations.open_browser(
            settings.config_get("REPO_URL"),
            "issue","new",
            body=f"My viewser version is {version('viewser')}"
        )




import datetime
import io
from typing import Optional, Dict, Any

import click
from viewser import settings
from viewser.settings import defaults
from . import operations, formatting

@click.group(name="queryset", short_help="queryset_operations related to querysets")
@click.pass_obj
def cli(ctx_obj: Dict[str, Any]):
    ctx_obj["operations"] = operations.QuerysetOperations(
            settings.QUERYSET_URL,
            defaults.default_error_handler(),
            settings.QUERYSET_MAX_RETRIES,
            )
    ctx_obj["table_formatter"] = formatting.QuerysetTableFormatter()
    ctx_obj["detail_formatter"] = formatting.QuerysetDetailFormatter()

@cli.command(name="fetch", short_help="fetch data for a queryset")
@click.argument("name")
@click.argument("out-file", type=click.File("wb"))
@click.option("-s","--start-date", type=click.DateTime())
@click.option("-e","--end-date", type=click.DateTime())
@click.pass_obj
def queryset_fetch(
        ctx_obj: Dict[str, Any],
        name:       str,
        out_file:   io.BufferedWriter,
        start_date: Optional[datetime.datetime],
        end_date:   Optional[datetime.datetime]):
    """
    Fetch data for a queryset named NAME from ViEWS cloud and save it to OUT_FILE
    """
    ctx_obj["operations"].fetch(name, out_file)

@cli.command(name="list", short_help="show a list of available querysets")
@click.pass_obj
def queryset_list(ctx_obj: Dict[str, Any]):
    """
    Show a list of available querysets.
    """
    ctx_obj["operations"].list().then(ctx_obj["table_formatter"].formatted).then(click.echo)

@cli.command(name="show", short_help="show details about a queryset")
@click.argument("name", type=str)
@click.pass_obj
def queryset_show(ctx_obj: Dict[str, Any], name: str):
    """
    Show detailed information about a queryset
    """
    ctx_obj["operations"].show(name).then(ctx_obj["detail_formatter"].formatted).then(click.echo).then(click.echo)

@cli.command(name="delete", short_help="delete a queryset")
@click.confirmation_option(prompt="Delete queryset?")
@click.argument("name", type=str)
@click.pass_obj
def queryset_delete(ctx_obj: Dict[str, Any], name: str):
    """
    Delete a queryset.
    """
    ctx_obj["operations"].delete(name)
    click.echo(f"Deleted {name}")



import sys
import io
import click
import json
import tabulate
from viewser import settings

@click.group(name="config", short_help="configure viewser")
def cli():
    """
    Configure viewser
    """

@cli.command(name="interactive", short_help="interactively configure viewser")
def config_interactive():
    """
    Interactively configure viewser (useful for first-time config)
    """
    settings.configure_interactively()
    click.echo("All done!")

@cli.command(name="set", short_help="set a configuration value")
@click.argument("name", type=str)
@click.argument("value", type=str, default=1)
@click.option("--override/--no-override",default=True)
def config_set(name: str, value: str, override: bool): # pylint: disable=redefined-builtin
    """
    Set a configuration value.
    """
    overrides = True
    try:
        settings.config.get(name)
    except settings.exceptions.ConfigurationError:
        overrides = False

    if not override and overrides:
        click.echo(f"Setting {name} already set, override not specified (see --help)")
        return

    settings.config.set(name,value)
    click.echo(f"{name}: {value}")

@cli.command(name="get", short_help="get a configuration value")
@click.argument("name", type=str)
def config_get(name: str):
    """
    Get a configuration value.
    """
    value = settings.config.get(name)
    if value:
        click.echo(f"{name}: {value}")
    else:
        click.echo(f"{name} not set")
        sys.exit(1)

@cli.command(name="reset", short_help="reset default config values")
@click.confirmation_option(prompt="Reset config?")
def config_reset():
    """
    Reset config, writing default values over current values.
    """
    settings.config.load(settings.static.DEFAULT_SETTINGS, overwrite = True)
    click.echo("Config file reset")

@cli.command(name="list", short_help="show all configuration settings")
def config_list():
    """
    Show all current configuration values
    """
    click.echo(tabulate.tabulate(settings.config.list().items()))

@cli.command(name="unset", short_help="unset a configuration value")
@click.confirmation_option(prompt="Unset config key?")
@click.argument("name", type=str)
def config_unset(name: str):
    """
    Unset a configuration value, removing its entry from the config file.
    """
    current_value = settings.config.get(name)
    settings.config.unset(name)
    click.echo(f"Unset {name} (was {current_value})")

@cli.command(name = "dump", short_help = "dump current config as JSON")
def config_dump():
    click.echo(json.dumps(settings.config.list()))

@cli.command(name = "load", short_help = "load configuration from a JSON file")
@click.confirmation_option(prompt="Load config file? (overwrites current config values)")
@click.argument("file", type = click.File("r"))
def config_load(file: io.BufferedReader):
    settings.config.load(json.load(file), overwrite = True)




import click

@click.group(name = "logs", short_help = "commands for viewing and streaming logs")
def cli():
    pass


@cli.command(name = "list", short_help = "show available log streams")
def list_logs():
    pass

@cli.command(name = "stream", short_help = "open a log stream")
def stream_log():
    pass

@cli.command(name = "get", short_help = "download a log")
def get_log():
    pass



"""
Compatibility
"""
from typing import Optional
from datetime import date
import pandas as pd
from viewser import settings
from viewser.settings import defaults
from viewser.commands.queryset import operations

def fetch(queryset_name: str, start_date: Optional[date] = None, end_date: Optional[date] = None) -> Optional[pd.DataFrame]:
    """
    fetch
    =====

    parameters:
        queryset_name (str)

    returns:
        pandas.DataFrame

    Fetch a queryset

    """
    return operations.QuerysetOperations(
            settings.REMOTE_URL,
            defaults.default_error_handler(),
            settings.QUERYSET_MAX_RETRIES,
            ).fetch(queryset_name).maybe(None, lambda x:x)



from viewser import error_handling
from . import config

def default_error_handler():
    return error_handling.ErrorDumper([
        error_handling.FileErrorHandler(config.get("ERROR_DUMP_DIRECTORY")),
        error_handling.StreamHandler()])



"""
queryset_operations
===================

"""
import sys
import time
from typing import Optional
from io import BytesIO, BufferedWriter
from datetime import date
import logging
from pyarrow.lib import ArrowInvalid
import pandas as pd
import requests
from toolz.functoolz import do, curry
from pymonad.either import Either, Left, Right
from pymonad.maybe import Just, Nothing, Maybe
from views_schema import viewser as viewser_schema
from views_schema import queryset_manager as queryset_schema
from viewser import remotes
from viewser.error_handling import errors, error_handling
from viewser.tui import animations

from . import queryset_list

logger = logging.getLogger(__name__)

response_json = lambda rsp: rsp.json()

class QuerysetOperations():

    def __init__(self,
            remote_url: str,
            error_handler: Optional[error_handling.ErrorDumper] = None,
            max_retries: int = sys.maxsize):
        self._remote_url = remote_url
        self._max_retries = max_retries
        self._error_handler = error_handler if error_handler else error_handling.ErrorDumper([])

    def fetch(self, queryset_name:str, out_file: Optional[BufferedWriter] = None, start_date: Optional[date] = None, end_date: Optional[date] = None) -> Maybe[pd.DataFrame]:
        """
        fetch
        =====

        parameters:
            queryset_name (str): Name of the queryset to fetch
            out_file (BufferedWriter): File to write queryset to

        returns:
            Maybe[pandas.DataFrame]: Dataframe corresponding to queryset (if query succeeds)

        """
        f = self._fetch(
                self._max_retries,
                self._remote_url,
                queryset_name,
                start_date, end_date)
        if out_file is not None:
            f.then(curry(do, lambda data: data.to_parquet(out_file)))

        return f.either(self._error_handler.dump, Just)

    def list(self) -> Maybe[queryset_list.QuerysetList]:
        """
        list
        ====

        returns:
            Optional[List[str]]: Returns a list of queryset name if operation succeeds.

        """
        return (self._request("GET", remotes.status_checks, "querysets")
            .then(lambda r: r.json())
            .then(lambda d: queryset_list.QuerysetList(**d))
            .either(self._error_handler.dump, Just))

    def show(self, name: str) -> Maybe[queryset_schema.DetailQueryset]:
        """
        show
        ====

        parameters:
            name (str): Name of the queryset to show

        returns:
            Optional[viewser_schema.queryset_manager.DetailQueryset]: Returns queryset model if successful.

        """
        return (self._request("GET", remotes.status_checks, f"querysets/{name}")
            .then(lambda r: r.json())
            .then(lambda d: queryset_schema.DetailQueryset(**d))
            .either(self._error_handler.dump, Just))

    def publish(self, queryset: queryset_schema.Queryset, overwrite: bool = True) -> Maybe[requests.Response]:
        (self._request(
                "POST",
                remotes.status_checks,
                "querysets",
                parameters = Just({"overwrite":overwrite}),
                data = Just(queryset.dict()))
            .either(self._error_handler.dump, Just))

    def delete(self, name: str) -> Maybe[requests.Response]:
        (self._request( "DELETE",
                remotes.status_checks,
                f"querysets/{name}",
                )
            .either(self._error_handler.dump, Just))

    def _request(self, method: str, checks, path, **kwargs) -> Either[viewser_schema.Dump, requests.Response]:
        return remotes.request(self._remote_url, method, checks, path, **kwargs)

    def _deserialize(self, response: requests.Response) -> Either[viewser_schema.Dump, pd.DataFrame]:
        if response.status_code == 202:
            # No data yet
            return Right(None)
        else:
            try:
                return Right(pd.read_parquet(BytesIO(response.content)))
            except (OSError, ArrowInvalid):
                return Left(errors.deserialization_error(response))

    def _fetch(
            self,
            max_retries : int,
            base_url: str, name: str,
            start_date: Optional[date] = None, end_date: Optional[date] = None
            ) -> Either[viewser_schema.Dump, pd.DataFrame]:
        """
        _fetch
        ======

        Fetches queryset located at {base_url}/querysets/data/{name}

        Args:
            base_url(str)
            name(str)
            start_date(Optional[str]): Only fetch data after start_date
            start_date(Optional[str]): Only fetch data before end_date

        Returns:
            Either[errors.Dump, pd.DataFrame]

        """
        start_date, end_date = [date.strftime("%Y-%m-%d") if date else None for date in (start_date, end_date)]

        checks = [
                    remotes.check_4xx,
                    remotes.check_error,
                    remotes.check_404,
                 ]

        parameters = {
                k:v for k,v in {"start_date":start_date, "end_date":end_date}.items() if v is not None
                }
        parameters = Just(parameters) if len(parameters) > 0 else Nothing
        path = f"data/{name}"

        retries = 0
        anim    = animations.LineAnimation()
        data    = Right(None)

        while (data.is_right() and data.value is None):
            if retries > 0:
                time.sleep(1)

            anim.print_next()

            data = (remotes.request(base_url, "GET", checks, path, parameters = parameters)
                .then(self._deserialize))

            if retries > max_retries:
                data = Left(errors.max_retries())

            retries += 1

        anim.end()

        return data




import pickle
import io
from typing import Dict
import click
from views_storage import key_value_store
from viewser.storage import model_object

@click.group(name = "model", short_help = "management of model objects")
@click.pass_context
def cli(ctx: click.Context):
    ctx.obj["model-object-storage"]          = model_object.ModelObjectStorage()
    ctx.obj["model-object-metadata-storage"] = model_object.ModelMetadataStorage()

@cli.command(name = "list", short_help = "show a list of model objects")
@click.pass_obj
def list_models(obj: Dict[str, key_value_store.KeyValueStore]):
    click.echo("\n".join(obj["model-object-storage"].list()))

@cli.command(name = "download", short_help = "download a model object to a file")
@click.argument("name", type = str)
@click.argument("file", type = click.File("wb"))
@click.pass_obj
def download_model(
        obj: Dict[str, key_value_store.KeyValueStore],
        name: str,
        file: io.BufferedWriter):
    try:
        pickle.dump(obj["model-object-storage"].read(name), file)
    except FileNotFoundError:
        click.echo(f"No model named {name}")

@cli.command(name = "inspect", short_help = "display metadata for a model object")
@click.argument("name", type = str)
@click.pass_obj
def inspect_model(obj: Dict[str, key_value_store.KeyValueStore], name: str):
    try:
        click.echo(obj["model-object-metadata-storage"].read(name))
    except FileNotFoundError:
        click.echo(f"No model named {name}")





from copy import deepcopy

def deepcopy_self(fn):
    """
    deepcopy_self
    =============

    Decorator that allows models to return copies of themselves, allowing for
    safe method chaining.
    """
    def inner(self, *args,**kwargs):
        return fn(deepcopy(self), *args, **kwargs)
    return inner



from typing import Callable, Dict, Union, Any, Optional
import strconv
from sqlalchemy.orm import Session
from . import models, exceptions

class ConfigResolver():

    def __init__(self, sessionmaker: Callable[[], Session]):
        self._Session = sessionmaker

    def get(self, key: str, default: Optional[Any] = None) -> Union[str, int, float]:
        with self._Session() as session:
            try:
                assert (setting := self._get(session, key)) is not None
            except AssertionError:
                if default is None:
                    raise exceptions.ConfigurationError(f"Configuration setting {key} is not set")
                return default
            else:
                return strconv.convert(setting.value)

    def set(self, key, value) -> None:
        with self._Session() as session:
            return self._set(session, key, value)

    def unset(self, key) -> None:
        with self._Session() as session:
            return self._unset(session, key)

    def load(self, settings: Dict[str, Union[str, int, float]], overwrite: bool = False) -> None:
        with self._Session() as session:
            for key,value in settings.items():
                if not self._exists(session, key) or overwrite:
                    self._set(session, key, value)

    def list(self) -> Dict[str, Union[str, int, float]]:
        with self._Session() as session:
            settings = session.query(models.Setting).all()
            return {s.key: s.value for s in settings}

    def _get(self, session: Session, key: str) -> Union[str, int, float]:
        return session.query(models.Setting).get(key)

    def _exists(self, session: Session, key: str) -> bool:
        return session.query(models.Setting).get(key) is not None

    def _set(self, session: Session, key: str, value: Union[str, int, float]) -> None:
        if (existing := self._get(session,key)) is not None:
            existing.value = value
        else:
            setting = models.Setting(key = key, value = value)
            session.add(setting)
        session.commit()

    def _unset(self, session: Session, key: str) -> None:
        session.delete(session.query(models.Setting).get(key))
        session.commit()




from views_storage import key_value_store
from views_storage.backends import sftp
from views_storage.serializers import pickle, json
from viewser import settings

class ModelObjectStorage(key_value_store.KeyValueStore):
    def __init__(self):
        super().__init__(
                backend = sftp.Sftp(
                    host          = settings.config_get("MODEL_OBJECT_SFTP_HOSTNAME"),
                    port          = settings.config_get("MODEL_OBJECT_SFTP_PORT"),
                    user          = settings.config_get("MODEL_OBJECT_SFTP_USER"),
                    key_db_host   = settings.config_get("MODEL_OBJECT_KEY_DB_HOSTNAME"),
                    key_db_dbname = settings.config_get("MODEL_OBJECT_KEY_DB_DBNAME"),
                    folder        = "data/model_objects"),
                serializer = pickle.Pickle()
                )

    def list(self):
        raw = super().list()
        return raw.files

class ModelMetadataStorage(key_value_store.KeyValueStore):
    def __init__(self):
        super().__init__(
                backend = sftp.Sftp(
                    host          = settings.config_get("MODEL_OBJECT_SFTP_HOSTNAME"),
                    port          = settings.config_get("MODEL_OBJECT_SFTP_PORT"),
                    user          = settings.config_get("MODEL_OBJECT_SFTP_USER"),
                    key_db_host   = settings.config_get("MODEL_OBJECT_KEY_DB_HOSTNAME"),
                    key_db_dbname = settings.config_get("MODEL_OBJECT_KEY_DB_DBNAME"),
                    folder        = "data/metadata"),
                serializer = json.Json()
                )

    def list(self):
        raw = super().list()
        return raw.files



"""
transform
=========

These classes are used to provide an ergonomic API for end-users, via some
delegation and copying magic. The idea is just to allow users to do this:

    my_column.transform.my_namespace.my_transform(*args)

This returns a new column with the my_namespace.my_transform(*args) transform
appended to the list of column operations.

Not for human consumption.

"""
from views_schema import queryset_manager as schema


class Transform():
    """
    Transform
    =========

    parameters:
        column (viewser.Column): The column to add the transform to
        namespace (str): A prospective transform namespace name
        name (str): A prospective transform name

    Represents a prospective transform. Can be called to return the column with
    the transform operation appended.
    """
    def __init__(self, column, namespace, name):
        self.namespace = namespace
        self.name = name
        self.column = column

    def __call__(self, *arguments):
        op = schema.TransformOperation(
                name = self.namespace + "." + self.name,
                arguments = arguments)

        column = self.column.copy()
        column._add_trf(op)
        return column

class TransformNamespace():
    """
    TransformNamespace
    ==================

    parameters:
        column (viewser.Column): The column to add the transform to
        name (str): A prospective transform namespace name

    Represents a transform namespace

    """
    def __init__(self, column, name: str):
        self.name = name
        self.column = column

    def __getattribute__(self, name):
        if not name in ["column","name"]:
            return Transform(self.column, self.name, name)
        else:
            return super().__getattribute__(name)

class TransformNamespaces():
    """
    TransformNamespaces
    ===================

    parameters:
        column (viewser.Column): The column to add the transform to

    Represents the set of possible transform namespaces.
    """
    def __init__(self, column):
        self.column = column

    def __getattribute__(self, name):
        if not name == "column":
            return TransformNamespace(self.column, name)
        else:
            return super().__getattribute__(name)




from typing import List
import pydantic

class QuerysetList(pydantic.BaseModel):
    """
    This is a temp fix, alleviating the fact that the queryset manager module doesn't return the right data
    """
    querysets: List[str]




from views_schema import docs as schema

class WrappedViewsDoc:
    def __init__(self, doc: schema.ViewsDoc):
        self.doc = doc

    @property
    def name(self):
        return self.doc.entry.name

    @property
    def heading(self):
        return self.name.capitalize()

    @property
    def has_page(self):
        return self.doc.page is not None

    @property
    def description(self):
        return self.doc.page.content if self.has_page else "No description added"

    @property
    def author(self):
        return self.doc.page.author if self.has_page else None

    @property
    def date_updated(self):
        return str(self.doc.page.last_edited.date()) if self.has_page else None

    @property
    def sub_entries(self):
        return self.doc.entry.entries

    @property
    def data(self):
        return self.doc.entry.data

    def json(self):
        return self.doc.json()



from functools import partial
import itertools

print_no_newline = partial(print, end = "", flush = True)

class WaitingAnimation():
    CYCLE = ["-","\\", "|", "/"]
    def __init__(self, message: str = ""):
        self._animation_cycle = itertools.cycle(self.CYCLE)
        self._previous_output = ""
        self._message = message

    def _erase_prev(self):
        print_no_newline("\b" * len(self._previous_output))

    def print_next(self):
        to_print = self._message + " " + next(self._animation_cycle)
        self._erase_prev()
        self._previous_output = to_print
        print_no_newline(to_print)

    def end(self):
        self._erase_prev()

class LineAnimation(WaitingAnimation):
    CYCLE = [
        ".     ",
        " o    ",
        "  O   ",
        "   O  ",
        "    o ",
        "     .",
        ]




class HirearchicalDict(dict):
    def __setitem__(self, key, value):
        try:
            self[key]
        except KeyError:
            super().__setitem__(key, value)
        else:
            pass




import logging
from typing import Optional, Dict, Any

import tabulate
from views_schema import docs as schema

from viewser.tui.formatting import abc, styles
from . import wrapped_views_doc

logger = logging.getLogger(__name__)

DocumentationSection = abc.Section[wrapped_views_doc.WrappedViewsDoc]

class Title(DocumentationSection):
    def compile_output(self, model: wrapped_views_doc.WrappedViewsDoc) -> str:
        return "> "+ styles.bold(model.heading)

class Description(DocumentationSection):
    TITLE = "Description"

    def compile_output(self, model: wrapped_views_doc.WrappedViewsDoc) -> Optional[str]:
        if model.has_page:
            return "\n".join((
                    model.description,
                    " ",
                    model.author + "@" + model.date_updated,
                    " ",
                ))
        else:
            return None

class FunctionSection(DocumentationSection):
    TITLE = "Function description"

    def compile_output(self, model: wrapped_views_doc.WrappedViewsDoc) -> Optional[str]:
        try:
            signature = model.data["arguments"]

            arguments = ["- " + a["name"] + " (" + a["type"] + ")" for a in signature][1:]
            if arguments:
                sig_description = f"{model.heading} takes {len(arguments)} argument(s):"
                signature_lines = [sig_description] + arguments
            else:
                signature_lines = [f"{model.heading} takes no arguments"]

            if model.data["level_of_analysis"] == "any":
                type_note = "This transform can be applied to any kind of data."
            else:
                type_note = (
                        "This transform can only be applied to "
                        f"data of type {model.data['level_of_analysis']}"
                    )

            modelstring = model.data["docstring"] if model.data["docstring"] is not None else "..."

            return "\n".join((
                "Docstring:",
                "\""+modelstring+"\"",
                " ",
                *signature_lines,
                " ",
                type_note
            ))
        except KeyError:
            logger.warning("Failed to parse returned documentation")
            return None

class DocumentationTableSection(DocumentationSection):
    def str_str_dict(self, any_dict: Dict[Any,Any])-> Dict[str,str]:
        return {k:v for k,v in any_dict.items() if isinstance(k,str) and isinstance(v,str)}

    def compile_output(self, model: wrapped_views_doc.WrappedViewsDoc)-> str:
        data = []
        for entry in model.sub_entries:
            row = self.str_str_dict(entry.dict())
            row.update(self.str_str_dict(entry.data))
            data.append(row)
        return tabulate.tabulate(data, headers="keys", tablefmt = "pipe") + "\n"

class DocumentationFormatter(abc.Formatter[schema.ViewsDoc]):
    def formatted(self, model: schema.ViewsDoc)-> str:
        doc = wrapped_views_doc.WrappedViewsDoc(model)
        return super().formatted(doc)

class DocumentationDetailFormatter(DocumentationFormatter):
    SECTIONS = [
            Title,
            Description,
        ]

class DocumentationTableFormatter(DocumentationFormatter):
    SECTIONS = [
            Title,
            Description,
            DocumentationTableSection
        ]

class FunctionDetailFormatter(DocumentationFormatter):
    SECTIONS = [
            FunctionSection,
            Description
        ]




import tabulate
from views_schema import viewser as schema
from viewser.tui.formatting import abc
from . import models

class ErrorDumpFilesTable(abc.Section[models.ErrorDumpFiles]):
    TITLE = "Error dumps"

    def compile_output(self, model: models.ErrorDumpFiles):
        return tabulate.tabulate(
                [(i+1, f.name, str(f.size)+" bytes") for i, f in enumerate(model.files)],
                headers = ("filename", "size"),
                tablefmt = "pipe")

class DumpHeader(abc.Section[schema.Dump]):
    TITLE = "Error dump"

    def compile_output(self, model: schema.Dump)-> str:
        return (f"\"{model.title}\"\n"
            f"This error was posted at {model.timestamp} by {model.username}")

class MessagesSummary(abc.Section[schema.Dump]):
    TITLE = "Messages"

    def compile_output(self, model: schema.Dump)-> str:
        for message in model.messages:
            msg = message.message_type.name
            if message.message_type is schema.MessageType.DUMP:
                msg += f"\nSize: {len(message.content)}"
            else:
                msg += f"\n{message.content}"
        return msg

class DumpFileListFormatter(abc.Formatter[models.ErrorDumpFiles]):
    SECTIONS = [
            ErrorDumpFilesTable
        ]

class DumpFileErrorFormatter(abc.Formatter[schema.Dump]):
    SECTIONS = [
            DumpHeader,
            MessagesSummary,
        ]




from urllib import parse
import webbrowser

def open_browser(base,*args,**kwargs):
    querystring = parse.urlencode(kwargs)
    url = "/".join([base] + list(args))
    if querystring:
        url += "?"+querystring
    webbrowser.open(url)




from typing import List
from views_schema import viewser as schema
from . import abc

class ErrorSection(abc.Section[schema.Dump]):
    def _messages_of_kind(self, model: schema.Dump, kind: schema.MessageType) -> List[schema.Message]:
        return [m.content for m in model.messages if m.message_type is kind]

class ErrorMessages(ErrorSection):
    TITLE = "Error"

    def compile_output(self, model: schema.Dump) -> str:
        return "\n".join(self._messages_of_kind(model, schema.MessageType.MESSAGE))

class RecoveryHints(ErrorSection):
    TITLE = "Hints"

    def compile_output(self, model: schema.Dump) -> str:
        return "\n".join(self._messages_of_kind(model, schema.MessageType.HINT))

class ErrorFormatter(abc.Formatter[schema.Dump]):
    SECTIONS = [
        ErrorMessages,
        RecoveryHints,
        ]




import logging
from typing import Callable
import base64
import json
import datetime
import pydantic
import requests
from views_schema import viewser as schema

logger = logging.getLogger(__name__)

def response_as_json(response: requests.Response):
    """
    response_as_json
    ================

    Return a response in json form, used for dumping response contents.
    """
    try:
        content = response.content.decode()
    except UnicodeDecodeError:
        content = base64.b64encode(response.content).decode()

    return json.dumps({
            "url":         response.url,
            "http_status": response.status_code,
            "content":     content,
        })

def try_to_propagate(function: Callable[[requests.Response], schema.Dump])-> Callable[[requests.Response], schema.Dump]:
    """
    try_to_propagate
    ================

    Try to propagate an error dump from the remote, before running the local
    function to produce a fallback error.
    """
    def inner(response: requests.Response) -> schema.Dump:
        try:
            error = schema.Dump(**response.json())
            logger.debug("Propagating error from remote")
            return error 
        except (json.JSONDecodeError, pydantic.ValidationError):
            logger.debug(f"Failed to propagate error from response: \"{response.content[:25]}...\"")
            return function(response)
    return inner

def max_retries():
    return schema.Dump(
            title = "Max number of retries!",
            timestamp = datetime.datetime.now(),
            messages = [
                    schema.Message(
                            message_type = schema.MessageType.MESSAGE,
                            content = ("You reached the maximum number of "
                                "retries while waiting for data." )),
                    schema.Message(
                            message_type = schema.MessageType.HINT,
                            content = ("Try increasing the max. allowed number "
                                "of retries by running `viewser config RETRIES "
                                "{larger_number}"))
                ])

def connection_error(url: str):
    return schema.Dump(
            title = "Connection error!",
            timestamp = datetime.datetime.now(),
            messages = [
                    schema.Message(
                            message_type = schema.MessageType.MESSAGE,
                            content = (f"You failed to connect to {url}.")),
                    schema.Message(
                            message_type = schema.MessageType.HINT,
                            content = ("Is the REMOTE_URL correct? View the "
                                "currently set REMOTE_URL by running viewser "
                                "config show REMOTE_URL. Change it with viewser "
                                "config set REMOTE_URL."))
                ])

@try_to_propagate
def remote_error(response: requests.Response):
    return schema.Dump(
            title = "500 remote error",
            timestamp = datetime.datetime.now(),
            messages = [
                    schema.Message(
                        content = "Internal server error!"
                        ),

                    schema.Message(
                        message_type = schema.MessageType.HINT,
                        content = ("This kind of error is caused by something "
                            "going wrong on the server. Please contact and "
                            "administrator right away.")
                        ),

                    schema.Message(
                        message_type = schema.MessageType.DUMP,
                        content = response_as_json(response)
                        )
                ])

@try_to_propagate
def not_found_error(response: requests.Response):
    return schema.Dump(
            title = "404 not found",
            timestamp = datetime.datetime.now(),
            messages = [
                    schema.Message(
                        content = (f"The url {response.url} refers to a resource "
                        "that does not exist on the server.")),
                    schema.Message(
                        message_type = schema.MessageType.HINT,
                        content = "Did you mistype something?"),
                ])

@try_to_propagate
def client_error(response: requests.Response):
    return schema.Dump(
            title = f"{response.status_code} client error",
            timestamp = datetime.datetime.now(),
            messages = [
                    schema.Message(
                        content = ("The request was not accepted by the server, because it did not have the right format.")),
                    schema.Message(
                        message_type = schema.MessageType.HINT,
                        content = ("These messages might be caused by using an "
                        "old version of viewser that is not supported by the "
                        "currently running server version. Try updating viewser "
                        "with pip install --upgrade viewser. "
                        "It might also be caused by misspelling a column name, "
                        "a table name, or a transform name."
                        )),
                ])

def url_parsing_error(url: str):
    return schema.Dump(
            title = "Failed to parse URL",
            timestamp = datetime.datetime.now(),
            messages = [
                    schema.Message(
                        content = (f"Failed to parse url: {url}")),
                    schema.Message(
                        message_type = schema.MessageType.HINT,
                        content = ("This is indicative of an issue with "
                            "viewser. Contact an admin and copy-paste this error "
                            "message.")),
                ])

def request_exception(exception: requests.RequestException):
    return schema.Dump(
            title = f"{str(exception)}",
            timestamp = datetime.datetime.now(),
            messages = [
                    schema.Message(
                        content = ("Encountered an unhandled exception while making a request."))
                ])

def deserialization_error(response: requests.Response):
    return schema.Dump(
            title = "Deserialization error",
            timestamp = datetime.datetime.now(),
            messages = [
                    schema.Message(
                        content = ("Failed to deserialize data from response "
                            "content. The response had HTTP status "
                            f"{response.status_code}.")),
                    schema.Message(
                        message_type = schema.MessageType.DUMP,
                        content = (response_as_json(response))),
                ]
            )

def exists_error(resource_name: str):
    return schema.Dump(
            title = "Resource exists",
            timestamp = datetime.datetime.now(),
            messages = [
                    schema.Message(
                            content = f"Resource {resource_name} already exists"
                        ),
                ])



import os
from . import db, models, config_resolver, static

models.metadata.create_all(db.engine)
config = config_resolver.ConfigResolver(db.Session)
config.load(static.DEFAULT_SETTINGS, overwrite = False)

config_get = config.get

# =Remotes================================================

REMOTE_URL = config.get("REMOTE_URL")
QUERYSET_URL = os.path.join(REMOTE_URL, "querysets")

QUERYSET_MAX_RETRIES = config.get("QUERYSET_MAX_RETRIES")

FOO = config.get("bar", "baz")

# =Compatibility==========================================

def reset_defaults():
    config.load(static.DEFAULT_SETTINGS, overwrite = True)

def configure_interactively():
    for setting in static.REQUIRED_SETTINGS:
        value = None
        while not value:
            value = input(f"{setting} >> ")
        config.set(setting, value)




import os
import json
import click
from views_schema import viewser as schema
from viewser import settings
from . import formatting
from . import models

@click.group(name = "system", short_help = "administration commands")
def cli():
    """
    Admin commands.
    """

@cli.group(name = "error", short_help = "commands related to error dumps")
@click.pass_context
def error(ctx: click.Context):
    """
    Commands related to error dumps.
    """
    ctx.obj["errors"] = models.ErrorDumpFiles.from_dir(settings.ERROR_DUMP_DIRECTORY)

@error.command(name = "list", short_help = "show current error dumps")
@click.pass_context
def list_errors(ctx: click.Context):
    """
    Show a list of current error dumps.
    """
    click.echo(formatting.DumpFileListFormatter().formatted(
            ctx.obj["errors"]))

@error.command(name = "clear", short_help = "delete all current error dumps")
@click.confirmation_option(prompt = "Delete all error dumps?")
@click.pass_context
def delete_errors(ctx: click.Context):
    """
    Delete all current error dumps.
    """
    for file in ctx.obj["errors"].files:
        os.unlink(os.path.join(ctx.obj["errors"].directory, file.name))

@error.command(name = "show", short_help = "delete all current error dumps")
@click.argument("number", type = int, default = 1)
@click.pass_context
def show_error(ctx: click.Context, number: int):
    try:
        err = ctx.obj["errors"].files[number-1]
    except IndexError:
        click.echo(f"No such error ({number}). Try running viewser system errors list.")
    else:
        with open(os.path.join(ctx.obj["errors"].directory,err.name)) as f:
            error_model = schema.Dump(**json.load(f))
        click.echo(formatting.DumpFileErrorFormatter().formatted(error_model))




import os

CONFIG_DIR = os.path.abspath(os.path.expanduser("~/.views"))
CONFIG_DB_FILE = os.path.join(CONFIG_DIR, "settings.sqlite")

DEFAULT_SETTINGS = {
        "RETRY_FREQUENCY":                  5,
        "LOG_LEVEL":                        "INFO",
        "HANDSHAKE_PATH":                   "",
        "REPO_URL":                         "https://www.github.com/prio-data/viewser",
        "LATEST_KNOWN_VERSION":             "0.0.0",
        "NOTEBOOK_SERVER_IMAGE_REPOSITORY": "prio-data/viewserspace",
        "NOTEBOOK_SERVER_IMAGE_REGISTRY":   "viewsregistry.azurecr.io",
        "ERROR_DUMP_DIRECTORY":             os.path.join(CONFIG_DIR, "dumps"),
        "MODEL_OBJECT_SFTP_USER":           "predictions",
        "MODEL_OBJECT_SFTP_PORT":           22222,
        "MODEL_OBJECT_SFTP_HOSTNAME":       "hermes",
        "MODEL_OBJECT_KEY_DB_HOSTNAME":     "janus",
        "MODEL_OBJECT_KEY_DB_DBNAME":       "pred3_certs",
        "QUERYSET_MAX_RETRIES":             500,
        "QUERYSET_REMOTE_PATH":             "querysets",
        "REMOTE_URL":                       "http://0.0.0.0:4000"
    }

REQUIRED_SETTINGS = (
            "REMOTE_URL",
        )



from typing import Optional
from toolz.functoolz import curry
from pymonad.maybe import Just, Nothing
import click
from viewser.error_handling import exceptions
from viewser import settings
from . import operations

@click.group(name="notebooks", short_help="run ipython notebooks")
def cli():
    pass

@cli.command(name="run", short_help="start the viewserspace ipython notebook server")
@click.option("-r","--requirements-file", type = str, default = None, help = "Optional requirements file")
@click.option("-w","--work-dir", type = str, default = ".", help = "Directory to mount")
@click.option("-v","--use-version", type = str, default="latest", help="Image version to use")
@click.option("--browser/--no-browser", type = bool, default=True, help="Automatically open a browser window")
@click.option("--pull/--no-pull", type = bool, default=True, help="Check for environment updates")
def run_viewserspace(
        use_version: str,
        work_dir: str,
        browser: bool,
        pull: bool,
        requirements_file: Optional[str],
        ):

    requirements_file = Just(requirements_file) if requirements_file else Nothing
    port = operations.seek_port(operations.START_PORT)

    if not port:
        click.echo("No available ports")
    else:
        (operations.notebook_image(
                settings.config_get("NOTEBOOK_SERVER_IMAGE_REGISTRY"),
                settings.config_get("NOTEBOOK_SERVER_IMAGE_REPOSITORY"),
                use_version,
                pull)
            .then(curry(operations.run_notebook_server, port, work_dir, requirements_file))
            .either(
                curry(exceptions.exception_raiser,exceptions.ViewserspaceError),
                curry(operations.watch, browser, click.echo))
            )




import os
import logging

from views_schema import docs as schema
from views_schema.viewser import Dump
from pymonad.either import Either, Left, Right
from pymonad.maybe import Just
from pydantic import ValidationError
from requests.exceptions import JSONDecodeError

from viewser import remotes
from viewser.error_handling import errors

logger = logging.getLogger(__name__)

class DocumentationCrudOperations():
    """
    DocumentationCrudOperations
    ===========================

    parameters:
        base_url (str)
        path (str)

    Class for doing CRUD on documentation for views3
    """


    def __init__(self, base_url: str, path: str):
        self._base_url = os.path.join(base_url, path)

    def post(self, posted: schema.PostedDocumentationPage, name: str = None, overwrite: bool = False)-> Either[Dump, schema.ViewsDoc]:

        if name is None:
            name = posted.name

        if not overwrite:
            try:
                assert not self._exists(name)
            except AssertionError:
                return Left(errors.exists_error(name))

        return (remotes.request(self._base_url,"POST", remotes.status_checks, name, data = Just(posted.dict()))
            .then(lambda _: remotes.request(self._base_url, "GET", remotes.status_checks, name))
            .then(self._deserialize))

    def list(self) -> Either[Dump, schema.ViewsDoc]:
        return self.show("")

    def show(self, name: str) -> Either[Dump, schema.ViewsDoc]:
        return (remotes.request(self._base_url, "GET", remotes.status_checks, name)
            .then(self._deserialize))

    def _exists(self,name: str) -> Either[Dump, bool]:
        response = remotes.request(self._base_url, "GET", [], name)
        return response.is_right() and response.value.status_code == 200

    def _deserialize(self, response) -> Either[Dump, schema.ViewsDoc]:
        try:
            return Right(schema.ViewsDoc(**response.json()))
        except (JSONDecodeError, ValidationError):
            return Left(errors.deserialization_error(response))



"""
viewser.cli
===

This module defined the viewser cli, which is what is exposed when installing
the viewser package. The main entrypoint is viewser.cli.viewser. Command groups
are defined in separate modules under viewser.commands.
"""
from importlib.metadata import version
import logging
import click

from . import hirearchical_dict
from .tui.formatting import json_formatter
from .tui.formatting import errors as error_formatting
from .commands import queryset, config, documentation, help, notebooks, system, model

logger = logging.getLogger(__name__)

def print_version(ctx, _, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(version("viewser"))
    ctx.exit()

@click.group()
@click.option("--debug/--no-debug", default=False, help="Display debug logging messages")
@click.option("--json/--formatted", default=False, help="Output raw data, instead of formatted data")
@click.option("--version",
        is_flag = True, callback = print_version, expose_value = False, is_eager= True)
@click.pass_context
def viewser(ctx: click.Context, debug: bool, json: bool):
    ctx.obj = hirearchical_dict.HirearchicalDict()
    if json:
        ctx.obj["table_formatter"] = json_formatter.JsonFormatter()
        ctx.obj["detail_formatter"] = json_formatter.JsonFormatter()
    if debug:
        logging.getLogger().setLevel(logging.DEBUG)

    ctx.obj["error_dumper"] = error_formatting.ErrorFormatter()

viewser.add_command(queryset.cli)
viewser.add_command(config.cli)
viewser.add_command(documentation.tables_cli)
viewser.add_command(documentation.transforms_cli)
viewser.add_command(notebooks.cli)
viewser.add_command(system.cli)
viewser.add_command(help.cli)
viewser.add_command(model.cli)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['viewser',
 'viewser.commands',
 'viewser.commands.config',
 'viewser.commands.documentation',
 'viewser.commands.help',
 'viewser.commands.logs',
 'viewser.commands.model',
 'viewser.commands.notebooks',
 'viewser.commands.queryset',
 'viewser.commands.queryset.models',
 'viewser.commands.system',
 'viewser.error_handling',
 'viewser.settings',
 'viewser.storage',
 'viewser.tui',
 'viewser.tui.formatting']

package_data = \
{'': ['*']}

install_requires = \
['PyMonad>=2.4.0,<3.0.0',
 'click>=8.0.1,<9.0.0',
 'colorama>=0.4.4,<0.5.0',
 'crayons>=0.4.0,<0.5.0',
 'docker>=5.0.0,<6.0.0',
 'environs>=9.3.1,<10.0.0',
 'fitin>=0.2.0,<0.3.0',
 'pandas>=1.2.3,<2.0.0',
 'psutil>=5.8.0,<6.0.0',
 'pyarrow>=6.0.1,<7.0.0',
 'pydantic>=1.8.1,<2.0.0',
 'requests>=2.27.1,<3.0.0',
 'strconv>=0.4.2,<0.5.0',
 'tabulate>=0.8.9,<0.9.0',
 'toml>=0.10.2,<0.11.0',
 'toolz>=0.11.1,<0.12.0',
 'views-schema==2.0.0',
 'views-storage>=0.3.0,<0.4.0']

entry_points = \
{'console_scripts': ['viewser = viewser.cli:viewser']}

setup_kwargs = {
    'name': 'viewser',
    'version': '5.6.3',
    'description': '',
    'long_description': None,
    'author': 'peder2911',
    'author_email': 'pglandsverk@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from typing import Dict, Any, List, Callable
import json
import logging
from urllib import parse
import requests
from requests import Response
import pydantic
from pymonad.either import Left, Right, Either
from pymonad.maybe import Nothing, Maybe
from toolz.functoolz import curry, compose
from views_schema import viewser as schema

from .error_handling import errors

logger = logging.getLogger(__name__)

def make_request(
        url: str,
        method: str = "GET",
        json_data: Maybe[Dict[str,Any]] = Nothing) -> Either[schema.Dump, Response]:
    """
    make request

    Make a request to url, with optional data.

    Args:
        url(str): The url to request
        method(str): The method to use

        json_data(Maybe[Dict[str,Any]]): Optional JSON data to pass with request

    Returns:
        Either[schema.Dump, Response]
    """
    logger.debug(f"Making {method} request to {url}")
    request_args = [method, url]
    request_kwargs: Dict[str, Any] = {"headers":{}}

    def update_kwargs(kwargs, data):
        kwargs.update({"data":data})
        kwargs["headers"].update({"Content-Type":"application/json"})
        return kwargs

    request_kwargs = (json_data
        .then(json.dumps)
        .maybe(request_kwargs, curry(update_kwargs, request_kwargs))
        )

    try:
        response = requests.request(*request_args,**request_kwargs)
    except requests.exceptions.ConnectionError:
        return Left(errors.connection_error(url))
    except requests.RequestException as re:
        return Left(errors.request_exception(re))

    return Right(response)

def response_check(
        check_fn: Callable[[Response],bool],
        exc_fn: Callable[[Response], schema.Dump],
        response: Response) -> Either[schema.Dump, requests.Response]:
    """
    response_check

    Returns a function that can be used to check a
    response. The resulting check function returns
    either a Left containing an exception, or a Right
    containing the response.

    Args:
        check_fn: A callable that takes a response, and returns a bool
        exc_fn: A callable that takes a response, and returns an exception
        response: Response

    Returns:
        Either[Exception, Response]
    """
    if not check_fn(response):
        return Left(exc_fn(response))
    else:
        return Right(response)

ResponseCheck = Callable[[Either[schema.Dump, requests.Response]], Either[schema.Dump, requests.Response]]

def check_error(response: Response) -> Either[schema.Dump, requests.Response]:
    """
    check_error

    Checks whether a response returned a status 5xx, which signifies that the
    request resulted in an error on the remote server.
    """
    logger.debug(f"Checking if {response.url} resulted in an error")
    return response_check(lambda rsp: str(rsp.status_code)[0] != "5",
            errors.remote_error,
            response
        )

def check_404(response: Response) -> Either[schema.Dump, requests.Response]:
    """
    check_404

    Checks whether a 4xx error occurred while making the request.
    """
    return response_check(
            lambda rsp: rsp.status_code != 404,
            errors.not_found_error,
            response,
            )

def check_4xx(response: Response) -> Either[schema.Dump, requests.Response]:
    """
    check_4xx

    Checks whether a 4xx error occurred while making the request.
    """
    return response_check(
            lambda rsp: str(rsp.status_code)[0] != "4",
            errors.client_error,
            response,
            )

status_checks = [
        check_4xx,
        check_404,
        check_error,
    ]

def make_url(base_url: str, path: str, parameters = Nothing)-> Either[Exception, str]:
    """
    make_url

    Make an url, by combining a base url, a path, and an
    optional dictionary of parameters.

    Args:
        base_url(str): Protocol and host, ex http://foo.com
        path(str): Path to add to the base, ex. /foo/bar
        parameters(Maybe[Dict[str,str]]): Optional dict of parameters

    Returns:
        str
    """
    try:
        raw = base_url.strip("/") + "/" + path

        url = pydantic_validate(raw, pydantic.HttpUrl)

        with_pstring = (parameters
            .then(parse.urlencode)
            .maybe(url, lambda x: url + "?" + x)
            )
        return Right(with_pstring)

    except pydantic.ValidationError:
        return Left(errors.url_parsing_error(raw))
def compose_checks(
        checks: List[ResponseCheck]
        )-> Callable[[Either[schema.Dump, requests.Response]], Either[schema.Dump, requests.Response]]:
    make_then = lambda fn: lambda x: x.then(fn)
    return compose(*[make_then(fn) for fn in checks])

def request(
        base_url: str,
        method: str,
        checks: List[ResponseCheck],
        path: str,
        parameters: Maybe[Dict[str,str]] = Nothing,
        data: Maybe[Dict[str,Any]] = Nothing
        ) -> Either[schema.Dump, Response]:
    """
    request

    Make a request, returning either a Right(response) or a Left(exception).

    Args:
        base_url(str): The base url
        method(str): The method to use (GET, POST, etc)
        checks(ResponseCheck): A list of checks (see response_check)

        path(str): The path to request

        parameters(Maybe[Dict[str,str]]): An optional dict of query-string parameters
        data(Maybe[Dict[str,Any]]): An optional JSON dict to send to the server

    Returns:
        Either[schema.Dump, Response]
    """

    data.then(str).then(lambda r: f"POSTing {r}").then(logger.debug)

    url = make_url(base_url, path, parameters)
    response = url.then(curry(make_request, method = method, json_data = data))
    return compose_checks(checks)(response)

def pydantic_validate(val, type):
    class mdl(pydantic.BaseModel):
        x: type
    return mdl(x=val).x




import tabulate
from views_schema import queryset_manager as schema
from viewser.tui.formatting import abc
from . import queryset_list

class QuerysetListTable(abc.Section[queryset_list.QuerysetList]):
    TITLE = "Querysets"
    def compile_output(self, model: queryset_list.QuerysetList) -> str:
        return tabulate.tabulate([(m,) for m in model.querysets], ("name",), tablefmt = "pipe")

class QuerysetTableFormatter(abc.Formatter[queryset_list.QuerysetList]):
    SECTIONS = [
            QuerysetListTable,
        ]

class QuerysetName(abc.Section[schema.DetailQueryset]):
    def compile_output(self, model: schema.DetailQueryset):
        return f"{model.name} at {model.loa}"

class QuerysetDescription(abc.Section[schema.DetailQueryset]):
    TITLE = "Description"

    def compile_output(self, model: schema.DetailQueryset):
        return model.description

class QuerysetSummary(abc.Section[schema.DetailQueryset]):
    TITLE = "Summary"

    def compile_output(self, model: schema.DetailQueryset):
        return f"Queryset has {len(model.operations)} operations"

class QuerysetDetailFormatter(abc.Formatter[schema.DetailQueryset]):
    SECTIONS = [
            QuerysetName,
            QuerysetDescription,
            QuerysetSummary,
        ]




from contextlib import contextmanager
import re
from typing import Optional, List, Generic, TypeVar, Callable
from abc import ABC, abstractmethod
import click
import colorama

T = TypeVar("T")

class Section(ABC, Generic[T]):
    TITLE: str = ""

    def add_section(self, model: T, formatter: "Formatter")-> None:
        compiled = self.compile_output(model)
        if compiled:
            with formatter.text_in_section(self.TITLE):
                formatter.write(formatter.indented(compiled))

    @abstractmethod
    def compile_output(self, model: T) -> Optional[str]:
        pass

class Formatter(click.HelpFormatter, ABC, Generic[T]):
    _max_data_len = 150
    _divider_length = 32
    _divider_char = "-"
    SEPARATOR = "{{SEP}}"

    @property
    @abstractmethod
    def SECTIONS(self)-> List[Callable[[], Section[T]]]:
        pass

    def __init__(self):
        self.SECTIONS = [s() for s in self.SECTIONS]
        colorama.init()
        super().__init__()

    def indented(self, str) -> str:
        return "\n".join([(" "*self.current_indent)+ ln for ln in str.split("\n")])

    def divider(self) -> str:
        return self.indented(self.SEPARATOR)

    def write_heading(self, heading: str):
        self.write(self.indented(
                colorama.Style.BRIGHT + heading.capitalize() + colorama.Style.RESET_ALL + "\n"
            ))
        self.write("\n")

    @contextmanager
    def text_in_section(self, name: Optional[str] = None):
        if name:
            self.write(self.indented(">> "+ name)+"\n\n")
        try:
            yield
        finally:
            self.write("\n"+self.indented(self.SEPARATOR)+"\n")

    def _format(self, model: T):
        for s in self.SECTIONS:
            s.add_section(model, self)

    def formatted(self, model: T) -> str:
        self.write("\n")
        self.indent()
        self._format(model)
        rendered = self.getvalue()
        longest_line = max([len(l) for l in rendered.split("\n")])
        return re.sub(self.SEPARATOR,"-"*longest_line,rendered)




from views_schema import queryset_manager as schema
from . import transform, util

class Column():
    """
    Column
    ======

    parameters:
        name (str): Name to give to the resulting column
        from_table (str): Name of table from which to fetch this column
        from_column (str): Name of column in table to fetch

    description
    -----------

    Columns are used when defining querysets. They are defined as lists of
    operations, following a database retrieval operation, that represent
    various transformations of the data.

    help
    ----

    These CLI commands give lists of available columns and transforms:

    Tables and columns:
    ```
    viewser tables list / viewser tables show $TABLE_NAME
    ```

    Transforms:
    ```
    viewser transforms list / viewser transforms show $TRANSFORM_NAME
    ```
    """
    def __init__(self, name: str, from_table: str, from_column: str, _inject = None ):

        self._name = name
        self._from_table = from_table
        self._from_column = from_column

        inject = [] if not _inject else _inject

        self.operations = [
                schema.RenameOperation(arguments=[name]),
                *inject,
                schema.DatabaseOperation(
                    name = from_table+"."+from_column, arguments = ["values"]
                    )
            ]

        self.namespaces = transform.TransformNamespaces(self)

    def _add_trf(self, trf):
        self.operations.insert(1, trf)

    def aggregate(self, aggregation: str):
        """
        aggregate
        =========

        parameters:
            aggregation (str)

        Choice of aggregation function, if column must be aggregated on
        retrieval: If a column is fetched to a higher level of analysis, such
        as from subnational to national level.

        Valid aggregation functions are:
            - sum
            - max
            - avg
            - min
        """
        cp = self.copy()
        cp.set_aggregation(aggregation)
        return cp 

    def set_aggregation(self, agg):
        self.operations[-1].arguments[0] = agg

    @property
    def aggregation(self):
        return self.operations[-1].arguments[0]

    @property
    def transform(self):
        """
        transform
        =========

        Add a transform to the column. This is done by issuing this kind of
        call, which returns a transformed column:

        ```
        column.transform.namespace.function(*args)
        ```

        `namespace` and `function` must correspond to a valid namespace and a
        valid function name in that namespace. For a list of functions and
        namespaces, run this command on the command line:

        ```
        viewser transforms list
        ```
        """
        return self.namespaces

    def copy(self):
        to_inject = self.operations[1:len(self.operations)-1]
        new = Column(self._name, self._from_table, self._from_column, _inject = to_inject)
        new.set_aggregation(self.aggregation)
        return new





import datetime
import sqlalchemy as sa
from sqlalchemy.ext.declarative import declarative_base

metadata = sa.MetaData()
Base = declarative_base(metadata = metadata)

class Setting(Base):
    __tablename__ = "setting"
    key      = sa.Column(sa.String(), primary_key = True)
    value    = sa.Column(sa.String())
    modified = sa.Column(sa.DateTime(), default = datetime.datetime.now)



import logging
from views_schema import queryset_manager as schema
from viewser.commands.queryset.operations import QuerysetOperations
from viewser import settings
from viewser.settings import defaults
from . import column, util

logger = logging.getLogger(__name__)
queryset_operations = QuerysetOperations(
        settings.QUERYSET_URL,
        defaults.default_error_handler())

class Queryset(schema.Queryset):
    """
    Queryset
    ========

    parameters:
        name (str): Unique name
        at (str): Level of analysis

    example:
        queryset = (Queryset("my-queryset", "country_month")
            .with_column(Column("a", "table-a","column-in-a")
                .aggregate("max")
                .transform.functions.my_function())
            .with_column(Column("b", "table-b", "column-in-b"))

        dataset = queryset.publish().fetch()

    description
    -----------

    A queryset is a data definition that can be used to fetch dynamically
    defined data from views3. Querysets are collections of Column objects,
    combined with a level of analysis definition, that define what data is
    fetched.

    Most methods can be chained, allowing for concise, readable queryset
    definitions.
    """
    def __init__(self, name, at):
        super().__init__(name = name, loa = at, operations = [])

    @util.deepcopy_self
    def with_column(self, col: column.Column):
        """
        with_column
        ===========

        parameters:
            col (viewser.Column)
        returns:
            self

        Add a column to the queryset
        """
        self.operations.append(col.operations)
        return self

    @util.deepcopy_self
    def describe(self, description: str):
        """
        describe
        ========

        parameters:
            description (str): Long-form description of the queryset

        returns:
            self
        """
        self.description = description
        return self

    @util.deepcopy_self
    def with_theme(self, theme):
        """
        with_theme
        ==========

        parameters:
            theme (str): A theme-name to associate with the queryset

        returns:
            self

        Associate the queryset with a theme, which can be any semantically
        meaningful string.
        """
        self.themes.append(theme)
        return self

    def publish(self, *args, **kwargs):
        """
        publish
        =======

        returns:
            self

        Publish the queryset to ViEWS 3, making it available to yourself and
        other users.
        """
        logger.info(f"Publishing queryset {self.name}")
        queryset_operations.publish(self, *args, **kwargs)
        return self

    def fetch(self, *args, **kwargs):
        """
        fetch
        =====

        returns:
            pandas.DataFrame

        Fetch the dataset corresponding to this queryset in its current state.
        Requires a self.push first.
        """
        logger.info(f"Fetching queryset {self.name}")
        dataset = queryset_operations.fetch(self.name, *args, **kwargs).maybe(None, lambda x:x)
        return dataset
